# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         Interval.py
# Purpose:      Pymuco classes
#
# Authors:      German Margon
#
# Copyright:    Copyright © 2022-2023 German Margon. All rights reserved.
# License:      BSD 3-Clause License, see LICENSE
# ------------------------------------------------------------------------------
from enum import IntEnum

from Enharmonic import Enharmonic
from NoteFrequencyConverter import NoteFrequencyConverter
from NoteMapping import NoteMapping
from ScientificPitchNotation import ScientificPitchNotation


class IntervalName(IntEnum):
    """
    This is a Python class called "IntervalName" that inherits from the
    IntEnum class. It is used to represent the names of different musical
    intervals. The class defines several class variables, each with a name and
    a value in semitones. For example, the UNISON variable has a value of 0,
    and the MINOR_SECOND variable has a value of 1.
    """

    UNISON = 0
    MINOR_SECOND = 1
    MAJOR_SECOND = 2
    MINOR_THIRD = 3
    MAJOR_THIRD = 4
    PERFECT_FOURTH = 5
    TRITONE = 6
    PERFECT_FIFTH = 7
    MINOR_SIXTH = 8
    MAJOR_SIXTH = 9
    MINOR_SEVENTH = 10
    MAJOR_SEVENTH = 11
    OCTAVE = 12
    MINOR_NINTH = 13
    MAJOR_NINTH = 14
    MINOR_TENTH = 15
    MAJOR_TENTH = 16
    PERFECT_ELEVENTH = 17
    DIMINISHED_TWELFTH = 18
    AUGMENTED_ELEVENTH = 19
    PERFECT_TWELFTH = 20
    MINOR_THIRTEENTH = 21
    MAJOR_THIRTEENTH = 22
    MINOR_FOURTEENTH = 23
    MAJOR_FOURTEENTH = 24
    DOUBLE_OCTAVE = 25


class Interval:
    """
    Represents an interval between two notes.

    Attributes
    ----------
    _low_note : ScientificPitchNotation
        The lower note of the interval.
    _high_note : ScientificPitchNotation
        The higher note of the interval.
    _semitones : int
        The number of semitones between the two notes.

    Methods
    -------
    low_note() -> ScientificPitchNotation:
        Returns the lower note of the interval.
    high_note() -> ScientificPitchNotation:
        Returns the higher note of the interval.
    get_interval_name() -> str:
        Returns the name of the interval.
    get_interval_direction() -> bool:
        Returns whether the interval is ascending or descending.
    get_interval_inversion() -> str:
        Returns the name of the interval after inverting the notes.
    """

    def __init__(
        self,
        low_note: ScientificPitchNotation,
        high_note: ScientificPitchNotation,
    ):
        """
        Initialize an Interval object with the given lower and higher notes.

        Parameters
        ----------
        low_note : ScientificPitchNotation
            The lower note of the interval.
        high_note : ScientificPitchNotation
            The higher note of the interval.

        Attributes
        ----------
        _low_note : ScientificPitchNotation
            The lower note of the interval.
        _high_note : ScientificPitchNotation
            The higher note of the interval.
        _semitones : int
            The number of semitones between the two notes.

        Returns
        -------
        None
        """
        self._low_note = low_note
        self._high_note = high_note
        self._semitones = self._get_semitones()

    @property
    def low_note(self):
        return self._low_note

    @property
    def high_note(self):
        return self._high_note

    def get_interval_name(self):
        """
        Get the name of the interval.

        Returns
        -------
        str
            The name of the interval between the lower and higher notes of this
            Interval object.

        Raises
        ------
        None
            This method does not raise any exceptions.
        """
        return IntervalName(self._semitones).name

    def get_interval_direction(self):
        """
        Determine the direction of the interval between the lower and higher
        notes of this Interval object.

        Returns
        -------
        bool
            True if the interval is ascending, False if the interval is
            descending.

        Raises
        ------
        None
        """
        return self._verify_interval_direction()

    def get_interval_inversion(self):
        """
        Returns the name of the inverted interval between the lower and higher
        notes of this Interval object.

        Parameters
        ----------
        None

        Returns
        -------
        str
            The name of the inverted interval as a string.

        Raises
        ------
        ValueError
            If the interval direction is neither ascending nor descending, or
            if the interval is greater than an octave.

        Notes
        -----
        If the interval direction is ascending, the lower note is raised by one
        octave and becomes the new higher note, while the higher note becomes
        the new lower note. If the interval direction is descending, the higher
        note is lowered by one octave and becomes the new lower note, while the
        lower note becomes the new higher note. The name of the inverted
        interval is determined by the number of semitones between the inverted
        notes.
        """
        interval_direction = self._verify_interval_direction()
        if interval_direction is True:
            self._low_note.octave += 1
            self._low_note, self._high_note = self._high_note, self._low_note
            semitones = self._get_semitones()
            interval_name = IntervalName(semitones)
            return interval_name.name.replace("_", " ")
        elif interval_direction is False:
            self._low_note.octave -= 1
            self._low_note, self._high_note = self._high_note, self._low_note
            semitones = self._get_semitones()
            interval_name = IntervalName(semitones)
            return interval_name.name.replace("_", " ")
        else:
            raise ValueError("Invalid interval.")

    def _get_semitones(self):
        """
        Calculates the number of semitones between the lower and higher notes
        of this Interval object.

        Args:
            None

        Returns:
            int: The number of semitones between the lower and higher notes as
            an integer.

        Raises:
            ValueError: If the interval between the notes is greater than an
            octave.
        """
        # Get note name mapping
        note_name_mapping = NoteMapping().get_note_mapping()
        # Get octave and pitch of lower and higher notes
        octave1 = self._low_note.octave
        octave2 = self._high_note.octave
        pitch1 = self._low_note.pitch
        pitch2 = self._high_note.pitch
        # Calculate the number of semitones
        semitones1 = note_name_mapping[pitch1]
        semitones2 = note_name_mapping[pitch2]
        semitones = (semitones2 - semitones1) + (octave2 - octave1) * 12
        if semitones < 0:
            return abs(semitones)
        # Raise an error if the interval is greater than an octave
        if semitones > 12:
            raise ValueError("Interval too large")
        else:
            return semitones

    def _verify_interval_direction(self):
        """
        This method verifies the direction of the interval between the low note
        and the high note of the Interval object. It returns a boolean value
        indicating whether the interval is ascending or descending.

        Returns:
        bool: True if the interval is ascending, False if the interval is
        descending.
        """
        note1_frequency = (
            NoteFrequencyConverter().get_frequency_from_note_name(
                self._low_note
            )
        )
        note2_frequency = (
            NoteFrequencyConverter().get_frequency_from_note_name(
                self._high_note
            )
        )
        if note1_frequency > note2_frequency:
            # Descending
            return False
        else:
            # Ascending
            return True


class IntervalCalculator:
    """
    The IntervalCalculator class provides methods for calculating intervals
    between notes, given a root note and an interval name.

    Parameters
    ----------
    None

    Attributes
    ----------
    None

    Methods
    -------
    calculate_ascending_interval(interval_name: IntervalName, root_note:
    ScientificPitchNotation) -> str:
        Calculates the interval between a root note and a note that is a given
        interval above the root note. Returns the name of the interval in
        string format.

    calculate_descending_interval(interval_name: IntervalName,
    root_note: ScientificPitchNotation) -> str:
        Calculates the interval between a root note and a note that is a given
        interval below the root note. Returns the name of the interval in
        string format.

    Notes
    -----
    The class uses a note name mapping to convert note names to numerical
    values, which are used to calculate the interval. The class also uses the
    Enharmonic class to calculate enharmonic equivalents of intervals.

    Examples
    --------
    To calculate an ascending interval:

    >>> ic = IntervalCalculator()
    >>> interval_name = IntervalName('M3')
    >>> root_note = ScientificPitchNotation('C4')
    >>> ic.calculate_ascending_interval(interval_name, root_note)
    'E4'

    To calculate a descending interval:

    >>> ic = IntervalCalculator()
    >>> interval_name = IntervalName('P5')
    >>> root_note = ScientificPitchNotation('A4')
    >>> ic.calculate_descending_interval(interval_name, root_note)
    'D4'

    """

    _NOTE_NAME_MAPPING = NoteMapping().get_note_mapping()

    def calculate_ascending_interval(
        self, interval_name: IntervalName, root_note: ScientificPitchNotation
    ):
        """
        Calculate the pitch of a note ascending a certain interval from a given
        root note.

        Parameters
        interval_name : IntervalName
        An instance of the IntervalName class representing the desired
        interval.
        root_note : ScientificPitchNotation
        An instance of the ScientificPitchNotation class representing the root
        note.

        Returns
        str
        The pitch of the note ascending the specified interval from the root
        note, as a string.

        Raises
        None

        Notes
        The method first retrieves the index of the root note in the mapping of
        note names to their corresponding indices. It then calculates the index
        of the note that is the specified interval above the root note, taking
        into account that the interval should wrap around to the beginning of
        the octave if necessary.

        Next, it retrieves the pitch of the note at the calculated index from
        the mapping of note names to their corresponding pitches. If the note
        name has a double sharp (x) or double flat (𝄫) accidental, the method
        replaces the note name with the appropriate enharmonic equivalent.

        Finally, if the root note and calculated note have different accidental
        markings, the method replaces the note name with the appropriate
        enharmonic equivalent. The method returns the resulting note name as a
        string.
        """
        root_note_str = str(root_note.pitch)
        root_note_index = self._NOTE_NAME_MAPPING[root_note_str]
        ascending_interval_index = (root_note_index + interval_name.value) % 12

        ascending_interval = [
            k
            for k, v in self._NOTE_NAME_MAPPING.items()
            if v == ascending_interval_index
        ][0]

        if len(ascending_interval) > 1 and ascending_interval[1] in ["x"]:
            ascending_interval = [
                k
                for k, v in self._NOTE_NAME_MAPPING.items()
                if v == ascending_interval_index
            ][1]
        elif len(ascending_interval) > 1 and ascending_interval[1] in ["𝄫"]:
            ascending_interval = [
                k
                for k, v in self._NOTE_NAME_MAPPING.items()
                if v == ascending_interval_index
            ][-1]

        # enharmonic from here
        if len(ascending_interval) == 1 and len(root_note_str) == 1:
            return ascending_interval

        elif ascending_interval[-1] == "#" and root_note_str[-1] == "𝄫":
            ascending_interval = (
                Enharmonic(ScientificPitchNotation(ascending_interval, 4))
                .get_enharmonic()
                .pitch
            )

        elif ascending_interval[-1] == "#" and root_note_str[-1] == "b":
            ascending_interval = (
                Enharmonic(ScientificPitchNotation(ascending_interval, 4))
                .get_enharmonic()
                .pitch
            )

        elif ascending_interval[-1] == "b" and root_note_str[-1] == "#":
            ascending_interval = (
                Enharmonic(ScientificPitchNotation(ascending_interval, 4))
                .get_enharmonic()
                .pitch
            )

        return ascending_interval

    def calculate_descending_interval(
        self, interval_name: IntervalName, root_note: ScientificPitchNotation
    ):
        """
        Calculate the descending interval between a root note and a given
        interval name and return the resulting note as a string in scientific
        pitch notation.

        Parameters
        ----------
        interval_name : IntervalName
            An instance of the IntervalName class representing the desired
            interval.
        root_note : ScientificPitchNotation
            An instance of the ScientificPitchNotation class representing the
            root note for the interval calculation.

        Returns
        -------
        str
            The pitch of the note descending the specified interval from the
            root note, as a string in scientific pitch notation.

        Raises
        ------
        ValueError
            If the root_note is not a valid scientific pitch notation.

        Notes
        -----
        The method first converts the root_note to a string and looks up its
        corresponding index in the _NOTE_NAME_MAPPING dictionary. It then
        calculates the index of the desired descending interval using the given
        interval_name and the index of the root_note. The method retrieves the
        string representation of the resulting note by searching the
        _NOTE_NAME_MAPPING dictionary for the note with the calculated index.

        The method then applies enharmonic equivalents to the resulting note if
        needed. If the note has only one character, the method immediately
        returns it. Otherwise, it checks the last character of the root_note
        and the resulting note to see if they need to be enharmonically
        adjusted. If an adjustment is necessary, the method creates a
        ScientificPitchNotation object for the resulting note, applies the
        get_enharmonic method of the Enharmonic class to it, and converts the
        resulting object back to a string in ScientificPitchNotation format.

        Examples
        --------
        >>> from music_theory.interval_name import IntervalName
        >>> from music_theory.scientific_pitch_notation import
        ScientificPitchNotation
        >>> from music_theory.interval_calculator import IntervalCalculator
        >>> root_note = ScientificPitchNotation.from_pitch_string('C4')
        >>> interval_name = IntervalName.MAJOR_THIRD
        >>> interval_calculator = IntervalCalculator()
        >>> interval_calculator.calculate_descending_interval(interval_name,
        root_note)
        'A3'"""

        root_note_str = str(root_note.pitch)
        root_note_index = self._NOTE_NAME_MAPPING[root_note_str]
        descending_interval_index = (
            root_note_index - interval_name.value
        ) % 12

        descending_interval = [
            k
            for k, v in self._NOTE_NAME_MAPPING.items()
            if v == descending_interval_index
        ][0]

        if len(descending_interval) > 1 and descending_interval[1] in ["x"]:
            descending_interval = [
                k
                for k, v in self._NOTE_NAME_MAPPING.items()
                if v == descending_interval_index
            ][1]
        elif len(descending_interval) > 1 and descending_interval[1] in ["𝄫"]:
            descending_interval = [
                k
                for k, v in self._NOTE_NAME_MAPPING.items()
                if v == descending_interval_index
            ][-1]

        # enharmonic from here
        if len(descending_interval) == 1:
            return descending_interval

        elif len(root_note_str) == 1 and descending_interval[-1] == "#":
            descending_interval = (
                Enharmonic(ScientificPitchNotation(descending_interval, 4))
                .get_enharmonic()
                .pitch
            )

        elif descending_interval[-1] == "#" and root_note_str[-1] == "𝄫":
            descending_interval = (
                Enharmonic(ScientificPitchNotation(descending_interval, 4))
                .get_enharmonic()
                .pitch
            )

        elif descending_interval[-1] == "#" and root_note_str[-1] == "b":
            descending_interval = (
                Enharmonic(ScientificPitchNotation(descending_interval, 4))
                .get_enharmonic()
                .pitch
            )

        elif descending_interval[-1] == "b" and root_note_str[-1] == "#":
            descending_interval = (
                Enharmonic(ScientificPitchNotation(descending_interval, 4))
                .get_enharmonic()
                .pitch
            )

        return descending_interval



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         AudioConverter.py
# Purpose:      Pymuco classes
#
# Authors:      German Margon
#
# Copyright:    Copyright © 2022-2023 German Margon. All rights reserved.
# License:      BSD 3-Clause License, see LICENSE
# ------------------------------------------------------------------------------
import math
import os
import re
import secrets
import struct
import tempfile
import wave
from typing import List, Tuple, Union

from MusicComputationNotation import MusicComputationNotation
from NoteDuration import NoteDuration, NoteDurationMapping
from NoteFrequencyConverter import NoteFrequencyConverter
from ScientificPitchNotation import ScientificPitchNotation


class AudioConverter:
    """
    Class Name: AudioConverter

    Attributes:
    -----------

    _sample_rate (int): The sample rate to use for the audio file.
    _name (str): A randomly generated name to use for the output file.
    _file_path (str): The directory path to write the output file to.
    _note_duration_mapping (NoteDurationMapping): A NoteDurationMapping object
    used to convert duration strings to their corresponding values in seconds.
    _note_frequency_converter (NoteFrequencyConverter):
    A NoteFrequencyConverter object used to convert note names to their
    corresponding frequencies.
    _full_file_path (str): The full path of the output file, including the
    directory and file name.

    Methods:
    --------

    init(self, sample_rate=44100): Initializes a new instance of the
    AudioConverter class with the specified sample rate.
    convert_to_audio(self, mcn: MusicComputationNotation) -> str: Converts a
    MusicComputationNotation object to an audio file and returns the file path.
    _validate_file_path(self): Checks that the file path for writing audio
    samples is valid and can be accessed.
    _create_sample(self, mcn: List[Tuple[str, Union[str, float]]]): Creates an
    audio sample from a list of musical notes and their durations. The input is
    a list of tuples, with each tuple containing a note name and its duration
    in string or float format. The output is a concatenated sample list of the
    notes.
    """

    def __init__(self, sample_rate=44100):
        """
        Initializes a new instance of the AudioConverter class with the
        specified sample rate.

        Args:
            sample_rate (int, optional): The sample rate to use for the audio
            file. Defaults to 44100.

        """
        self._sample_rate = sample_rate
        self._sample_list = []
        self._name = secrets.token_hex(8)
        self._file_path = os.path.join(tempfile.gettempdir(), "")
        self._note_duration_mapping = NoteDurationMapping()
        self._note_frequency_converter = NoteFrequencyConverter()
        self._full_file_path = self._file_path + self._name

    def convert_to_audio(self, mcn: MusicComputationNotation):
        """
        Convert a MusicComputationNotation object to an audio file.

        The method takes a MusicComputationNotation object containing a
        sequence of notes and durations, generates audio samples for each note
        using a sine wave generator, and writes the samples to an audio file in
        WAV format. The file is written to the directory specified by the
        _file_path attribute of the object.

        Parameters
        ----------
        mcn : MusicComputationNotation
            A MusicComputationNotation object containing a sequence of notes
            and durations to be converted to audio.

        Returns
        -------
        str
            The file path of the created audio file.

        Raises
        ------
        TypeError
            If the argument is not a MusicComputationNotation object.
        ValueError
            If the file path specified in the object is invalid or cannot be
            accessed for reading and writing.

        Examples
        --------
        # Convert a valid MusicComputationNotation object to audio
        >>> instance = MyClass()
        >>> mcn = MusicComputationNotation("C4 1/4 D4 1/4 E4 1/4 F4 1/4 G4 1/4
        A4 1/4 B4 1/4")
        >>> instance.convert_to_audio(mcn)
        '/path/to/output/audio.wav'

        # Convert an invalid argument to audio
        >>> instance = MyClass()
        >>> instance.convert_to_audio("invalid argument")
        Traceback (most recent call last):
        ...
        TypeError: mcn must be an instance of MusicComputationNotation

        # Convert to audio with an invalid file path
        >>> instance = MyClass()
        >>> instance._file_path = "/nonexistent/path"
        >>> mcn = MusicComputationNotation("C4 1/4")
        >>> instance.convert_to_audio(mcn)
        Traceback (most recent call last):
        ...
        ValueError: Invalid file path: /nonexistent/path

        """
        if not isinstance(mcn, MusicComputationNotation):
            raise TypeError(
                "mcn must be an instance of MusicComputationNotation"
            )
        self._validate_file_path()
        return self._create_sample(mcn.notation)

    def _validate_file_path(self):
        """
        Check that the file path for writing audio samples is valid and can be
        accessed.

        This method checks that the directory specified by the file path exists
        and can be read from and written to by the current user. If the
        directory does not exist or cannot be accessed, a ValueError is raised.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the file path specified in the object is invalid or cannot be
            accessed for reading and writing.

        Examples
        --------
        # Validate a valid file path
        >>> instance = MyClass()
        >>> instance._file_path = "/path/to/output"
        >>> instance._validate_file_path()

        # Validate an invalid file path
        >>> instance = MyClass()
        >>> instance._file_path = "/nonexistent/path"
        >>> instance._validate_file_path()
        Traceback (most recent call last):
        ...
        ValueError: Invalid file path: /nonexistent/path
        """
        if (
            not os.path.exists(self._file_path)
            or not os.access(self._file_path, os.R_OK)
            or not os.access(self._file_path, os.W_OK)
        ):
            raise ValueError(f"Invalid file path: {self._file_path}")

    def _create_sample(self, mcn: List[Tuple[str, Union[str, float]]]):
        """
        Create an audio sample from a list of musical notes and their
        durations.

        The method iterates through each note and duration in the list, and
        generates a list of audio samples for each note using the corresponding
        frequency and duration. The resulting audio samples are concatenated
        into a single sample list for the audio object, and the sample list is
        written to a WAV file with the specified name and sample rate.

        Args:
        -----
            mcn (List[Tuple[str, Union[str, float]]]): A list of tuples
                representing musical notes and their durations. Each tuple
                should contain a string representing the note name (in
                scientific pitch notation) and either a string or float
                representing the duration of the note in seconds or as a note
                duration string (e.g. "4n" for quarter note).

        Returns:
        --------
            None.

        Raises:
        -------
            ValueError: If the input list is empty or contains invalid note or
                duration values.
            OSError: If there is an error writing to the output file, such as
                if the file path is invalid or the output file cannot be opened
                for writing.

        Examples:
        ---------
            # Create an audio sample from a list of notes and durations
            >>> instance = MyClass()
            >>> instance._name = "test.wav"
            >>> instance._file_path = "/path/to/output"
            >>> instance._sample_rate = 44100
            >>> notes = [("A4", 0.5), ("C#5", "4n"), ("D5", 1.0)]
            >>> instance._create_sample(notes)

            # Check that the output file was created and contains audio
            >>> os.path.isfile(os.path.join(instance._file_path,
            instance._name))
            True
            >>> with wave.open(os.path.join(instance._file_path,
            instance._name), 'r') as file:
            ...     file.getnframes() > 0
            True
        """
        for note, duration in mcn:
            note_freq = self._get_frequency_note(note)
            duration = self._validate_duration(duration)
            self._generate_samples(note_freq, duration)
        return self._write_to_file()

    def _generate_samples(self, frequency: float, duration: float):
        """
        Generate a list of audio samples for a sine wave with the specified
        frequency and duration.

        Args:
        -----
            frequency (float): The frequency of the sine wave in Hz.
            duration (float): The duration of the sine wave in seconds.

        Returns:
        --------
            None.

        Raises:
        -------
            None.

        Examples:
        ---------
            Generate a list of samples for a sine wave with a frequency of
            440 Hz and duration of 1s:

            >>> instance = MyClass()
            >>> instance._sample_rate = 44100
            >>> instance.generate_samples(440, 1.0)

            Check that the sample list was populated with the expected number
            of samples:

            >>> len(instance._sample_list)
            44100
        """
        sample_rate = self._sample_rate
        num_samples = int(sample_rate * duration)
        constant = 2 * math.pi * frequency / sample_rate
        samples = [32767 * math.sin(constant * i) for i in range(num_samples)]
        self._sample_list.extend(samples)

    def _write_to_file(self):
        """
        Write the list of samples to a WAV file with the specified name and
        sample rate.

        Args:
        -----
        None.

        Returns:
        --------
        None.

        Raises:
        -------
        OSError: If there is an error writing to the output file, such as if
        the file path is invalid or the output file cannot be opened for
        writing.

        Examples:
        ---------
        # Write a list of samples to a WAV file
        >>> instance = MyClass()
        >>> instance._sample_list = [0.5, 0.1, -0.3, -0.7, 0.2]
        >>> instance._sample_rate = 44100
        >>> instance._name = "test.wav"
        >>> instance._file_path = "/path/to/output"
        >>> instance._write_to_file()
        # Check that the output file was created
        >>> os.path.isfile(os.path.join(instance._file_path, instance._name))
        True

        """
        num_samples = len(self._sample_list)
        file_path = os.path.join(self._file_path, self._name)
        with wave.open(file_path, "w") as file:
            file.setparams(
                (
                    1,
                    2,
                    self._sample_rate,
                    num_samples,
                    "NONE",
                    "not compressed",
                )
            )
            for sample in self._sample_list:
                file.writeframes(struct.pack("<h", int(sample)))

    def _get_frequency_note(self, note):
        """
        Get the frequency of a given note name in hertz.

        Args:
        -----
            note (str): A string representation of the note name, which can
            include the pitch name (e.g. "C", "G#", "Fb"), an optional
            accidental (e.g. "#", "b"), and an optional octave number
            (e.g. "4", "5", "-1").

        Returns:
        --------
            float: The frequency of the note in hertz, calculated using the
            Scientific Pitch Notation of the note and the standard tuning
            frequency.

        Raises:
        -------
            ValueError: If the input string does not match the expected format
            for a note name, or if the input note cannot be converted to a
            ScientificPitchNotation object.

        Examples:
        ---------
            # Get the frequency of a valid note name
            >>> instance = MyClass()
            >>> instance._get_frequency_note("C4")
            261.6255653005986

            # Get the frequency of a note name with an accidental
            >>> instance._get_frequency_note("G#3")
            207.65234878997256

            # Get the frequency of a note name with a flat accidental and
            # negative octave
            >>> instance._get_frequency_note("Fb-1")
            6.875

            # Invalid note name format
            >>> instance._get_frequency_note("invalid")  # Raises ValueError
            Traceback (most recent call last):
                ...
            ValueError: Invalid note name format: 'invalid'

            # Unable to convert input note to SPN
            >>> instance._get_frequency_note(None)  # Raises ValueError
            Traceback (most recent call last):
                ...
            ValueError: Unable to convert input note to SPN: 'None'
        """
        spn = self._convert_to_spn(note)
        if isinstance(spn, ScientificPitchNotation):
            return self._note_frequency_converter.get_frequency_from_note_name(
                spn
            )

    def _convert_to_spn(self, note):
        """
        Convert a note name in string format to a Scientific Pitch Notation
        object.

        Args:
        -----
            note: A string representation of the note name, which can include
            the pitch name (e.g. "C",
                "G#", "Fb"), an optional accidental (e.g. "#", "b"), and an
                optional octave number (e.g.
                "4", "5", "-1").

        Returns:
        --------
            A ScientificPitchNotation object representing the note, with the
            pitch name, accidental,
            and octave parsed from the input string.

        Raises:
        -------
            ValueError: If the input string does not match the expected format
            for a note name.

        Examples:
        ---------
            # Convert a valid note name to SPN
            >>> instance = MyClass()
            >>> spn = instance._convert_to_spn("C#4")
            >>> spn
            <ScientificPitchNotation C#4>

            # Convert a note name with no accidental or octave
            >>> instance._convert_to_spn("A")
            <ScientificPitchNotation A0>

            # Convert a note name with a flat accidental and negative octave
            >>> instance._convert_to_spn("Fb-1")
            <ScientificPitchNotation Fb-1>

            # Invalid note name format
            >>> instance._convert_to_spn("invalid")  # Raises ValueError
            Traceback (most recent call last):
                ...
            ValueError: Invalid note name format: 'invalid'
        """
        pitch_pattern = re.compile(r"\s*\d+\s*")
        octave_pattern = re.compile(r"[^0-9]")

        pitch = pitch_pattern.sub("", note)
        octave = octave_pattern.sub("", note)

        spn = ScientificPitchNotation(pitch, int(octave))
        return spn

    def _validate_duration(self, duration):
        """
        Check if a given duration is valid and convert it to a float value.

        Args:
        -----
            duration: The duration to be validated, which can be a string
            representation of a
                note duration (e.g. "quarter", "half", "eighth"), or a
                numerical value representing
                the duration in beats (e.g. 1.0 for a quarter note, 0.5 for an
                eighth note).

        Returns:
        --------
            A float value representing the duration in beats, if it is a valid
            duration. If the duration is a string representation of a note
            duration, it will be converted to its corresponding numerical value
            using the `_note_duration_mapping` dictionary. If the duration is a
            numerical value, it will be returned as is.

        Raises:
        -------
            ValueError: If the duration is not a valid note duration or
            numerical value.

        Examples:
        ---------
            # Check a valid duration as a string
            >>> instance = MyClass()
            >>> instance._validate_duration("half")
            2.0

            # Check a valid duration as a float
            >>> instance._validate_duration(0.25)
            0.25

            # Check an invalid duration
            >>> instance._validate_duration("quarter")  # Raises ValueError
            Traceback (most recent call last):
                ...
            ValueError: Invalid duration: 'quarter'
        """
        for (
            note,
            note_duration,
        ) in self._note_duration_mapping.get_note_duration():
            if note == duration or note_duration == duration:
                if isinstance(duration, int):
                    return float(duration)
                else:
                    return NoteDuration(duration).duration_length



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         MIDIUtils.py
# Purpose:      Pymuco classes
#
# Authors:      German Margon
#
# Copyright:    Copyright © 2022-2023 German Margon. All rights reserved.
# License:      BSD 3-Clause License, see LICENSE
# ------------------------------------------------------------------------------
from NoteMapping import NoteMapping
from ScientificPitchNotation import ScientificPitchNotation


class MIDINoteConverter:
    """
    The MIDINoteConverter class is used to convert note names represented in
    Scientific Pitch Notation to their corresponding MIDI note numbers, and
    vice versa.

    Attributes:
    -----------
    note_name_mapping (Dict[str, int]):
    A dictionary containing the mapping of note names to their corresponding
    MIDI pitch values.

    Methods:
    --------
    init(self):
    Initializes a MIDINoteConverter object by creating a dictionary of note
    names and their corresponding MIDI pitch values using the NoteMapping
    class. The resulting dictionary is stored as a private attribute of the
    object for use in other methods.

    note_name_to_midi_note(self, note_name: ScientificPitchNotation) -> int:
    Takes an instance of the ScientificPitchNotation class representing a note
    name, validates it, and returns its corresponding MIDI note number.

    midi_note_to_note_name(self, midi_note: int) -> str:
    Takes a MIDI note number, validates it, and returns the corresponding note
    name represented in Scientific Pitch Notation.

    _validate_note_name(self, note_name: ScientificPitchNotation) -> int:
    A helper method used to validate the note name input and return its
    corresponding MIDI pitch value. This method raises a ValueError if the
    input note name is not a valid note name.

    _validate_midi_note_range(self, midi_note: int) -> None:
    A helper method used to validate that the input MIDI note number is within
    the valid range of 0 to 127. This method raises a ValueError if the input
    MIDI note number is not within the valid range.

    The MIDINoteConverter class provides a convenient and reliable tool for
    converting between note names and MIDI note numbers in a Python project.
    It is designed to be used in conjunction with the ScientificPitchNotation
    class, and can be easily integrated into larger music software projects.
    """

    def __init__(self):
        """Initializes a MIDINoteConverter object.

        The method initializes the object by creating a dictionary of note
        names and their corresponding MIDI pitch values using the NoteMapping
        class. The resulting dictionary is stored as a private attribute of
        the object for use in other methods.

        Args:
            None
        Returns:
            None
        """
        self._note_name_mapping = NoteMapping().get_note_mapping()

    def note_name_to_midi_note(
        self, note_name: ScientificPitchNotation
    ) -> int:
        """Converts a note name represented in Scientific Pitch Notation to
        its corresponding MIDI note number.

        Args:
        -----
            note_name (ScientificPitchNotation): An instance of the
            ScientificPitchNotation class representing the note name to
            convert.

        Returns:
        --------
            int: The MIDI note number corresponding to the input note name.

        Raises:
        -------
            ValueError: If the input note_name argument is not an instance of
            the ScientificPitchNotation class or if the note name is not a
            valid note name.
        """
        if not isinstance(note_name, ScientificPitchNotation):
            raise ValueError(
                "Invalid argument: note_name must be an instance of"
                "ScientificPitchNotation"
            )

        midi_note = (
            int(note_name.octave) + 1
        ) * 12 + self._validate_note_name(note_name.pitch)
        self._validate_midi_note_range(midi_note)
        return midi_note

    def _validate_note_name(self, note_name: str) -> int:
        """Validates the input note name and returns its corresponding MIDI
        pitch value.

        Args:
        -----
            note_name (str): The note name to validate.

        Returns:
        --------
            int: The MIDI pitch value corresponding to the input note name.

        Raises:
        -------
            ValueError: If the input note name is not a valid note name.
        """
        if note_name not in self._note_name_mapping:
            raise ValueError(f"{note_name} is not a valid note name")
        return self._note_name_mapping[note_name]

    def midi_note_to_note_name(self, midi_note: int) -> str:
        """Converts a MIDI note number to its corresponding note name
        represented in Scientific Pitch Notation.

        Args:
        -----
            midi_note (int): The MIDI note number to convert.

        Returns:
        --------
            str: The note name represented in Scientific Pitch Notation
            corresponding to the input MIDI note number.

        Raises:
        -------
            ValueError: If the input MIDI note number is not within the valid
            range of 0 to 127.
        """
        self._validate_midi_note_range(midi_note)
        for note, value in self._note_name_mapping.items():
            if value == (midi_note % 12):
                return note + str(midi_note // 12 - 1)
        return "Invalid Note"

    def _validate_midi_note_range(self, midi_note):
        """Validates that the input MIDI note number is within the valid range
        of 0 to 127.

        Args:
        -----
            midi_note (int): The MIDI note number to validate.

        Raises:
        -------
            ValueError: If the input MIDI note number is not within the valid
            range of 0 to 127.
        """
        if not 0 <= midi_note <= 127:
            raise ValueError(f"{midi_note} is not a valid MIDI note number")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         Enharmonic.py
# Purpose:      Pymuco classes
#
# Authors:      German Margon
#
# Copyright:    Copyright © 2022-2023 German Margon. All rights reserved.
# License:      BSD 3-Clause License, see LICENSE
# ------------------------------------------------------------------------------
from EnharmonicMapping import EnharmonicMapping
from MusicData import DEFAULT_OCTAVE
from ScientificPitchNotation import ScientificPitchNotation


class Enharmonic:
    """
    Represents a musical note and provides functionality to determine its
    enharmonics.

    Attributes:
    -----------
        _note_name (ScientificPitchNotation): a ScientificPitchNotation object
        representing the note
        _is_enharmonic (bool): a boolean value indicating whether the note has
        an enharmonic
        _enharmonics (dict): a dictionary of notes and their enharmonics

    Methods:
    --------
        __init__(self, note):
            Initializes an Enharmonic object with a ScientificPitchNotation
            object representing a musical note. The _note_name instance
            variable is set to the input note, and the _is_enharmonic instance
            variable is set to True if the note has an enharmonic, and False
            otherwise. The _enharmonics instance variable is created by calling
            the enharmonic_mapping() method of the EnharmonicMapping class,
            which returns a dictionary of notes and their enharmonics.

        get_enharmonic(self):
            Returns a string representing the enharmonic of the note, or the
            original note name if it has no enharmonic.

        get_all_enharmonics(self):
            Returns a list of all the note names that are enharmonics of the
            original note, or a list containing only the original note name if
            it has no enharmonic.

        is_enharmonic(self):
            Returns True if the note has an enharmonic, and False otherwise.

        _calculate_enharmonic(self):
            Calculates the enharmonic of the note by looking up the _note_name
            instance variable in the _enharmonics instance variable. If the
            note is not found as a key, it looks up the value and returns the
            corresponding key.
    """

    def __init__(self, note: ScientificPitchNotation):
        """
        Initializes an Enharmonic object with a ScientificPitchNotation object
        representing a musical note.

        Args:
        -----
            note (ScientificPitchNotation): A ScientificPitchNotation object
            representing a musical note.

        Raises:
        -------
            ValueError: if the input note is not a valid scientific pitch
            notation.

        Attributes:
        -----------
            _note_name (ScientificPitchNotation): a ScientificPitchNotation
            object representing the note
            _is_enharmonic (bool): a boolean value indicating whether the note
            has an enharmonic
            _enharmonics (dict): a dictionary of notes and their enharmonics
        """
        self._enharmonics = EnharmonicMapping().enharmonic_mapping()
        self._note_name = note.pitch
        self._is_enharmonic = self.is_enharmonic()

    def get_enharmonic(self) -> str:
        """
        Returns a string representing the enharmonic of the note, or the
        original note name if it has no enharmonic.

        Returns:
        --------
            str: A string representing the enharmonic of the note, or the
            original note name if it has no enharmonic.
        """
        if self._is_enharmonic is False:
            return self._note_name
        self._note_name = self._calculate_enharmonic()
        return ScientificPitchNotation(self._note_name, DEFAULT_OCTAVE)

    def get_all_enharmonics(self) -> list:
        """
        Returns a list of all the note names that are enharmonics of the
        original note, or a list containing only the original note name if it
        has no enharmonic.

        Returns:
        --------
            list: A list of all the note names that are enharmonics of the
            original note, or a list containing only the original note name if
            it has no enharmonic.
        """
        if self._is_enharmonic is False:
            return [self._note_name]
        return [
            key if value == self._note_name else value
            for key, value in self._enharmonics.items()
            if key == self._note_name or value == self._note_name
        ]

    def is_enharmonic(self) -> bool:
        """
        Returns True if the note has an enharmonic, and False otherwise.

        Returns:
        --------
            bool: True if the note has an enharmonic, and False otherwise.
        """
        return self._note_name in self._enharmonics

    def _calculate_enharmonic(self) -> str:
        """
        Calculate the note name of this note's enharmonic equivalent.

        Returns:
        --------
        - A string representing the note name of this note's enharmonic
        equivalent, if it has one.
        - None, if this note has no enharmonic equivalent.

        Raises:
        -------
        - No exceptions are explicitly raised by this method.
        """
        return self._enharmonics.get(self._note_name) or next(
            (
                key
                for key, value in self._enharmonics.items()
                if value == self._note_name
            ),
            None,
        )
